"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.TilesetJsonCreator = void 0;
const fs_1 = __importDefault(require("fs"));
const path_1 = __importDefault(require("path"));
const core_1 = require("@gltf-transform/core");
const GltfTransform_1 = require("../contentProcessing/GltfTransform");
const PntsPointClouds_1 = require("../contentProcessing/pointClouds/PntsPointClouds");
const BufferedContentData_1 = require("../contentTypes/BufferedContentData");
const ContentDataTypes_1 = require("../contentTypes/ContentDataTypes");
const ContentDataTypeRegistry_1 = require("../contentTypes/ContentDataTypeRegistry");
const TileFormats_1 = require("../tileFormats/TileFormats");
const BoundingVolumes_1 = require("./BoundingVolumes");
const TileTableData_1 = require("../tileTableData/TileTableData");
const TileTableDataI3dm_1 = require("../tileTableData/TileTableDataI3dm");
const VecMath_1 = require("../tileTableData/VecMath");
const Loggers_1 = require("../logging/Loggers");
const logger = Loggers_1.Loggers.get("tilesetProcessing");
const DEFAULT_LEAF_GEOMETRIC_ERROR = 512;
const DEFAULT_TILESET_GEOMETRIC_ERROR = 4096;
/**
 * A class for creating `Tileset` JSON objects from tile content files.
 *
 * @internal
 */
class TilesetJsonCreator {
    // Implementation notes:
    //
    // To reduce ambiguities, the term "bounding box" refers to actual
    // `BoundingBox3D` instances. The term "bounding volume box" refers
    // to the 12-element number arrays that are the `boundingVolume.box`.
    //
    // Nearly all functions of this class are 'private', and some of
    // them make assumptions that only hold internally - namely, that
    // tiles do not have a `transform` property, or their bounding
    // volumes are actually `boundingVolume.box` bounding volumes.
    /**
     * Creates a tileset that uses the specified files as its
     * tile contents.
     *
     * Many details about the resulting tileset are intentionally
     * not specified.
     *
     * @param baseDir - The base directory against which the
     * content URIs are resolved
     * @param contentUris - The content URIs
     * @returns The tileset
     * @throws Error if content data could not be read
     */
    static async createTilesetFromContents(baseDir, contentUris) {
        const leafTiles = [];
        for (let i = 0; i < contentUris.length; i++) {
            const contentUri = contentUris[i];
            const leafTile = await TilesetJsonCreator.createLeafTileFromContent(baseDir, contentUri);
            if (leafTile) {
                leafTiles.push(leafTile);
            }
        }
        const tileset = TilesetJsonCreator.createTilesetFromLeafTiles(leafTiles);
        return tileset;
    }
    /**
     * Creates a leaf tile with the specified tile content.
     *
     * The content data will be loaded from the specified file, given
     * as `<baseDir>/<contentUri>`. The content data type will be
     * determined.
     *
     * If the content data type is one of the supported types (which
     * are unspecified for now), then its bounding volume will be
     * computed, and used as the bounding volume of the resulting tile.
     *
     * If the content data type is not supported, then a warning
     * will be printed and `undefined` will be returned.
     *
     * @param baseDir - The base directory against which the
     * content URI is resolved
     * @param contentUri - The content URI
     * @returns The leaf tile
     * @throws Error if content data could not be read
     */
    static async createLeafTileFromContent(baseDir, contentUri) {
        // Read the content data and determine its type
        const fileName = path_1.default.join(baseDir, contentUri);
        const data = fs_1.default.readFileSync(fileName);
        // Prepare the resolver for external GLBs in I3DM
        const externalGlbResolver = async (uri) => {
            const externalGlbUri = path_1.default.resolve(baseDir, uri);
            return fs_1.default.readFileSync(externalGlbUri);
        };
        const boundingBox = await TilesetJsonCreator.computeContentDataBoundingBox(contentUri, data, externalGlbResolver);
        if (!boundingBox) {
            logger.warn(`Content data type of ${contentUri} is not supported.`);
            return undefined;
        }
        logger.debug(`Bounding box for ${contentUri} is ` +
            `${boundingBox.min} ... ${boundingBox.max}`);
        const boundingVolumeBox = BoundingVolumes_1.BoundingVolumes.createBoundingVolumeBoxFromGltfBoundingBox(boundingBox);
        const boundingVolume = {
            box: boundingVolumeBox,
        };
        const geometricError = DEFAULT_LEAF_GEOMETRIC_ERROR;
        return TilesetJsonCreator.createLeafTile(boundingVolume, geometricError, contentUri);
    }
    /**
     * Computes the bounding box from the given content data.
     *
     * @param contentUri - The content URI
     * @param data - The content data
     * @param externalGlbResolver - The resolver for external GLBs in I3DMs
     * @returns The bounding box, or undefined if no bounding box could
     * be computed from the given content.
     * @throws Error if the I3DM referred to a GLB that could not be
     * resolved
     */
    static async computeContentDataBoundingBox(contentUri, data, externalGlbResolver) {
        const contentData = new BufferedContentData_1.BufferedContentData(contentUri, data);
        const contentDataType = await ContentDataTypeRegistry_1.ContentDataTypeRegistry.findContentDataType(contentData);
        if (contentDataType === ContentDataTypes_1.ContentDataTypes.CONTENT_TYPE_GLB) {
            return TilesetJsonCreator.computeBoundingBoxFromGlb(data);
        }
        else if (contentDataType === ContentDataTypes_1.ContentDataTypes.CONTENT_TYPE_PNTS) {
            return TilesetJsonCreator.computeBoundingBoxFromPnts(data);
        }
        else if (contentDataType === ContentDataTypes_1.ContentDataTypes.CONTENT_TYPE_B3DM) {
            return TilesetJsonCreator.computeBoundingBoxFromB3dm(data);
        }
        else if (contentDataType === ContentDataTypes_1.ContentDataTypes.CONTENT_TYPE_I3DM) {
            return TilesetJsonCreator.computeBoundingBoxFromI3dm(data, externalGlbResolver);
        }
        else if (contentDataType === ContentDataTypes_1.ContentDataTypes.CONTENT_TYPE_CMPT) {
            return TilesetJsonCreator.computeBoundingBoxFromCmpt(data, externalGlbResolver);
        }
        return undefined;
    }
    /**
     * Creates a tileset with the given leaf tiles.
     *
     * This assumes that the given tiles do not have a `transform`.
     *
     * If there is only one leaf tile, then this will become
     * the root of the tileset. Otherwise, the tileset root
     * will have the given tiles as its children.
     *
     * @param leafTiles - The leaf tiles
     * @returns The tileset
     */
    static createTilesetFromLeafTiles(leafTiles) {
        const tilesetGeometricError = DEFAULT_TILESET_GEOMETRIC_ERROR;
        let root = undefined;
        if (leafTiles.length === 1) {
            root = leafTiles[0];
        }
        else {
            root = TilesetJsonCreator.createParentTile(leafTiles);
        }
        root.refine = "ADD";
        const tileset = {
            asset: {
                version: "1.1",
            },
            geometricError: tilesetGeometricError,
            root: root,
        };
        return tileset;
    }
    /**
     * Creates a parent tile for the given child tiles.
     *
     * This assumes that the child tiles do not have a `transform`.
     *
     * It will compute the bounding volume box of the parent tile
     * from the bounding volume boxes of the children (using
     * `computeTilesUnionBoundingVolumeBox`), and a suitable (but unspecified)
     * geometric error for the parent tile.
     *
     * @param children - The children
     * @returns The parent tile
     */
    static createParentTile(children) {
        const geometricErrors = children.map((t) => t.geometricError);
        const maxGeometricError = Math.max(1, Math.max(...geometricErrors));
        const parentGeometricError = 2 * maxGeometricError;
        const parentBoundingVolumeBox = TilesetJsonCreator.computeTilesUnionBoundingVolumeBox(children);
        const tile = {
            boundingVolume: {
                box: parentBoundingVolumeBox,
            },
            geometricError: parentGeometricError,
            children: children,
        };
        return tile;
    }
    /**
     * Computes the `boundingVolume.box` that is the union of all bounding
     * volume boxes of the given tiles.
     *
     * This assumes that the child tiles do not have a `transform`.
     *
     * It also assumes that the input tile bounding volumes are actually
     * bounding boxes. Tiles that have different types of bounding
     * volumes will be ignored. If no tile has a bounding volume box,
     * then a unit cube bounding volume box will be returned.
     *
     * @param tiles - The input tiles
     * @returns The bounding volume box
     */
    static computeTilesUnionBoundingVolumeBox(tiles) {
        const boundingVolumeBoxes = tiles
            .map((t) => t.boundingVolume.box)
            .filter((b) => b !== undefined);
        return TilesetJsonCreator.computeUnionBoundingVolumeBox(boundingVolumeBoxes);
    }
    /**
     * Computes the union of the given boundingVolumeBoxes.
     *
     * If the given array is empty, then then a unit cube bounding
     * volume box will be returned.
     *
     * @param boundingVolumeBoxes - The bounding volume boxes
     * @returns The union volume box
     */
    static computeUnionBoundingVolumeBox(boundingVolumeBoxes) {
        let unionBoundingBox = undefined;
        for (const boundingVolumeBox of boundingVolumeBoxes) {
            const boundingBox = BoundingVolumes_1.BoundingVolumes.createBoundingBoxForBoundingVolumeBox(boundingVolumeBox);
            if (!unionBoundingBox) {
                unionBoundingBox = boundingBox;
            }
            else {
                unionBoundingBox = BoundingVolumes_1.BoundingVolumes.computeBoundingBoxUnion(unionBoundingBox, boundingBox);
            }
        }
        if (!unionBoundingBox) {
            unionBoundingBox = BoundingVolumes_1.BoundingVolumes.createUnitCubeBoundingBox();
        }
        logger.debug(`Union bounding box is ` +
            `${unionBoundingBox.min} ... ${unionBoundingBox.max}`);
        return BoundingVolumes_1.BoundingVolumes.createBoundingVolumeBoxFromBoundingBox(unionBoundingBox);
    }
    /**
     * Creates a leaf tile from the given data
     *
     * @param boundingVolume - The bounding volume
     * @param geometricError - The geometric error
     * @param contentUri - The content URI
     * @returns The tile
     */
    static createLeafTile(boundingVolume, geometricError, contentUri) {
        const tile = {
            boundingVolume: boundingVolume,
            geometricError: geometricError,
            content: {
                uri: contentUri,
            },
        };
        return tile;
    }
    /**
     * Computes the bounding box of the given PNTS data
     *
     * @param pntsBuffer - The PNTS data buffer
     * @returns A promise to the bounding box
     */
    static async computeBoundingBoxFromPnts(pntsBuffer) {
        // Read the tile data from the input data
        const tileData = TileFormats_1.TileFormats.readTileData(pntsBuffer);
        const batchTable = tileData.batchTable.json;
        const featureTable = tileData.featureTable.json;
        const featureTableBinary = tileData.featureTable.binary;
        // Create a `ReadablePointCloud` that allows accessing
        // the PNTS data
        const pntsPointCloud = await PntsPointClouds_1.PntsPointClouds.create(featureTable, featureTableBinary, batchTable);
        // Compute the minimum/maximum position
        const min = [Infinity, Infinity, Infinity];
        const max = [-Infinity, -Infinity, -Infinity];
        const globalPosition = pntsPointCloud.getGlobalPosition() ?? [0, 0, 0];
        const localPositions = pntsPointCloud.getPositions();
        for (const localPosition of localPositions) {
            const yupPosition = [
                localPosition[0] + globalPosition[0],
                localPosition[1] + globalPosition[1],
                localPosition[2] + globalPosition[2],
            ];
            const position = [
                yupPosition[0],
                yupPosition[2],
                -yupPosition[1],
            ];
            BoundingVolumes_1.BoundingVolumes.min(min, position, min);
            BoundingVolumes_1.BoundingVolumes.max(max, position, max);
        }
        const result = {
            min: min,
            max: max,
        };
        return result;
    }
    /**
     * Computes the bounding box of the given B3DM data
     *
     * @param b3dmBuffer - The B3DM data buffer
     * @returns A promise to the bounding box
     */
    static async computeBoundingBoxFromB3dm(b3dmBuffer) {
        // Compute the bounding box from the payload (GLB data)
        const tileData = TileFormats_1.TileFormats.readTileData(b3dmBuffer);
        const glbBuffer = tileData.payload;
        const gltfBoundngBox = await TilesetJsonCreator.computeBoundingBoxFromGlb(glbBuffer);
        // If the feature table defines an `RTC_CENTER`, then
        // translate the bounding box by this amount, taking
        // the y-up-to-z-up-conversion into account
        const featureTable = tileData.featureTable.json;
        if (featureTable.RTC_CENTER) {
            const featureTableBinary = tileData.featureTable.binary;
            const rtcCenter = TileTableData_1.TileTableData.obtainRtcCenter(featureTable.RTC_CENTER, featureTableBinary);
            const tx = rtcCenter[0];
            const ty = rtcCenter[2];
            const tz = -rtcCenter[1];
            const b3dmBoundingBox = BoundingVolumes_1.BoundingVolumes.translateBoundingBox(gltfBoundngBox, [tx, ty, tz]);
            return b3dmBoundingBox;
        }
        return gltfBoundngBox;
    }
    /**
     * Computes the bounding box of the given I3DM data
     *
     * @param i3dmBuffer - The I3DM data buffer
     * @param externalGlbResolver - The resolver for external GLB data from I3DMs
     * @returns A promise to the bounding box
     * @throws Error if the I3DM referred to a GLB that could not be
     * resolved
     */
    static async computeBoundingBoxFromI3dm(i3dmBuffer, externalGlbResolver) {
        // Obtain the GLB buffer for the tile data. With `gltfFormat===1`, it
        // is stored directly as the payload. Otherwise (with `gltfFormat===0`)
        // the payload is a URI that has to be resolved.
        const tileData = TileFormats_1.TileFormats.readTileData(i3dmBuffer);
        let glbBuffer = undefined;
        if (tileData.header.gltfFormat === 1) {
            glbBuffer = tileData.payload;
        }
        else {
            const glbUri = tileData.payload.toString().replace(/\0/g, "");
            glbBuffer = await externalGlbResolver(glbUri);
            if (!glbBuffer) {
                throw new Error(`Could not resolve external GLB from ${glbUri}`);
            }
        }
        // Compute the bounding box from the payload (GLB data)
        const gltfBoundingBox = await TilesetJsonCreator.computeBoundingBoxFromGlb(glbBuffer);
        const gltfCorners = BoundingVolumes_1.BoundingVolumes.computeBoundingBoxCorners(gltfBoundingBox);
        // Compute the instance matrices of the I3DM data
        const featureTable = tileData.featureTable.json;
        const featureTableBinary = tileData.featureTable.binary;
        const numInstances = featureTable.INSTANCES_LENGTH;
        const instanceMatrices = TileTableDataI3dm_1.TileTableDataI3dm.createInstanceMatrices(featureTable, featureTableBinary, numInstances);
        // Compute the minimum and maximum of the instance bounding boxes.
        // This is simply computed by transforming the corner points
        // of the glTF bounding box with the instancing transforms. This is
        // FAR from being the "tightest" bounding volume, but computing that
        // would require a robust OrientedBoundingBox::fromPoints function.
        const matrixZupToYup = VecMath_1.VecMath.createZupToYupPacked4();
        const matrixYupToZup = VecMath_1.VecMath.createYupToZupPacked4();
        const min = [Infinity, Infinity, Infinity];
        const max = [-Infinity, -Infinity, -Infinity];
        for (const matrix of instanceMatrices) {
            // Compute the matrix that first transforms the points from
            // the glTF coordinate system into Z-up, then applies the
            // I3DM transform, and then transforms them back to Y-up
            const gltfMatrix = VecMath_1.VecMath.multiplyAll4([
                matrixZupToYup,
                matrix,
                matrixYupToZup,
            ]);
            for (const gltfCorner of gltfCorners) {
                const transformedCorner = TilesetJsonCreator.transformPoint3D(gltfMatrix, gltfCorner);
                BoundingVolumes_1.BoundingVolumes.min(min, transformedCorner, min);
                BoundingVolumes_1.BoundingVolumes.max(max, transformedCorner, max);
            }
        }
        return {
            min: min,
            max: max,
        };
    }
    /**
     * Computes the bounding box of the given CMPT data
     *
     * @param cmptBuffer - The CMPT data buffer
     * @returns A promise to the bounding box
     */
    static async computeBoundingBoxFromCmpt(cmptBuffer, externalGlbResolver) {
        let unionBoundingBox = undefined;
        const compositeTileData = TileFormats_1.TileFormats.readCompositeTileData(cmptBuffer);
        const buffers = compositeTileData.innerTileBuffers;
        for (const buffer of buffers) {
            const innerBoundingBox = await TilesetJsonCreator.computeContentDataBoundingBox("[inner tile of CMPT]", buffer, externalGlbResolver);
            if (!innerBoundingBox) {
                continue;
            }
            logger.debug(`  Inner bounding box for CMPT is ` +
                `${innerBoundingBox.min} ... ${innerBoundingBox.max}`);
            if (!unionBoundingBox) {
                unionBoundingBox = innerBoundingBox;
            }
            else {
                unionBoundingBox = BoundingVolumes_1.BoundingVolumes.computeBoundingBoxUnion(unionBoundingBox, innerBoundingBox);
            }
        }
        if (!unionBoundingBox) {
            unionBoundingBox = BoundingVolumes_1.BoundingVolumes.createUnitCubeBoundingBox();
        }
        return unionBoundingBox;
    }
    /**
     * Transforms the given 3D point with the given 4x4 matrix, writes
     * the result into the given target, and returns it. If no target
     * is given, then a new point will be created and returned.
     *
     * @param matrix - The 4x4 matrix
     * @param point - The 3D point
     * @param target - The target
     * @returns The result
     */
    static transformPoint3D(matrix, point, target) {
        const px = point[0];
        const py = point[1];
        const pz = point[2];
        const x = matrix[0] * px + matrix[4] * py + matrix[8] * pz + matrix[12];
        const y = matrix[1] * px + matrix[5] * py + matrix[9] * pz + matrix[13];
        const z = matrix[2] * px + matrix[6] * py + matrix[10] * pz + matrix[14];
        if (!target) {
            return [x, y, z];
        }
        target[0] = x;
        target[1] = y;
        target[2] = z;
        return target;
    }
    /**
     * Computes the bounding box of the given glTF asset.
     *
     * This will compute the bounding box of the default scene
     * (or the first scene of the asset). If there is no scene,
     * then a warning will be printed, and a unit cube bounding
     * box will be returned.
     *
     * @param glbBuffer - The buffer containing GLB data
     * @returns A promise to the bounding box
     */
    static async computeBoundingBoxFromGlb(glbBuffer) {
        const io = await GltfTransform_1.GltfTransform.getIO();
        const document = await io.readBinary(glbBuffer);
        const root = document.getRoot();
        let scene = root.getDefaultScene();
        if (!scene) {
            const scenes = root.listScenes();
            if (scenes.length > 0) {
                scene = scenes[0];
            }
        }
        if (scene) {
            const bounds = (0, core_1.getBounds)(scene);
            return bounds;
        }
        logger.warn("No scenes found in glTF - using unit bounding box");
        return BoundingVolumes_1.BoundingVolumes.createUnitCubeBoundingBox();
    }
}
exports.TilesetJsonCreator = TilesetJsonCreator;
