import { Tileset } from "../structure/Tileset";
/**
 * A class for "upgrading" a tileset from a previous version to
 * a more recent version. The details of what that means exactly
 * are not (yet) specified.
 *
 * @internal
 */
export declare class TilesetUpgrader {
    /**
     * The options for the upgrade.
     */
    private readonly upgradeOptions;
    /**
     * The options that may be passed to `gltf-pipeline` when
     * GLB data in B3DM or I3DM is supposed to be upgraded.
     */
    private readonly gltfUpgradeOptions;
    /**
     * The tileset processor that will perform the actual upgrade
     */
    private tilesetProcessor;
    /**
     * Creates a new instance
     *
     * @param targetVersion - The target version - 1.0 or 1.1
     * @param gltfUpgradeOptions - Options that may be passed
     * to `gltf-pipeline` when GLB data in B3DM or I3DM is
     * supposed to be upgraded.
     * @throws DeveloperError If the version is neither 1.0 nor 1.1
     */
    constructor(targetVersion: string, gltfUpgradeOptions: any);
    /**
     * Creates pre-configured `TilesetUpgradeOptions` for the given
     * target version.
     *
     * @param version - The target version - 1.0 or 1.1
     * @returns The `TilesetUpgradeOptions`
     */
    private static optionsFor;
    /**
     * Upgrade the specified source tileset, and write it to the given
     * target.
     *
     * @param tilesetSourceName - The tileset source name
     * @param tilesetTargetName - The tileset target name
     * @param overwrite - Whether the target should be overwritten if
     * it already exists
     * @returns A promise that resolves when the process is finished
     * @throws TilesetError When the input could not be processed,
     * or when the output already exists and `overwrite` was `false`.
     */
    upgrade(tilesetSourceName: string, tilesetTargetName: string, overwrite: boolean): Promise<void>;
    /**
     * Perform the upgrade of the `Tileset` object, in place.
     *
     * @param tileset - The `Tileset` object
     */
    upgradeTileset(tileset: Tileset): Promise<void>;
    /**
     * Perform the upgrades of the tile contents, by processing all
     * content URIs with the `processContentUri`, and all content
     * values (files) with `processEntry`
     *
     * @param tilesetProcessor - The `BasicTilesetProcessor` that
     * will process the entries
     */
    private performContentUpgrades;
    /**
     * Process the given content URI.
     *
     * The given URI may either be an actual content URI, or a
     * template URI. Depending on the upgrade options, this
     * may change the file extension in the URI.
     *
     * @param uri - The URI
     * @returns The processed URI
     */
    private processContentUri;
    /**
     * Process the given tileset (content) entry, and return the result.
     *
     * @param sourceEntry - The source entry
     * @param type - The `ContentDataType` of the source entry
     * @returns The processed entry
     */
    private processEntry;
    /**
     * Process the given tileset (content) entry, and return the result.
     *
     * @param sourceEntry - The source entry
     * @param type - The `ContentDataType` of the source entry
     * @returns The processed entry
     */
    private processEntryUnchecked;
    /**
     * Process the given tileset (content) entry that contains PNTS,
     * and return the result.
     *
     * @param sourceEntry - The source entry
     * @returns The processed entry
     */
    private processEntryPnts;
    /**
     * Process the given tileset (content) entry that contains B3DM,
     * and return the result.
     *
     * @param sourceEntry - The source entry
     * @returns The processed entry
     */
    private processEntryB3dm;
    /**
     * Process the given tileset (content) entry that contains I3DM,
     * and return the result.
     *
     * @param sourceEntry - The source entry
     * @returns The processed entry
     */
    private processEntryI3dm;
    /**
     * Process the given tileset (content) entry that contains the
     * JSON of an external tileset, and return the result.
     *
     * @param sourceEntry - The source entry
     * @returns The processed entry
     */
    private processEntryTileset;
}
//# sourceMappingURL=TilesetUpgrader.d.ts.map